# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests resource config."""


import marshmallow as ma
from invenio_records_resources.resources import (
    RecordResourceConfig,
    SearchRequestArgsSchema,
)

from .fields import ReferenceString


#
# Request args
#
class RequestSearchRequestArgsSchema(SearchRequestArgsSchema):
    """Add parameter to parse tags."""

    created_by = ReferenceString()
    topic = ReferenceString()
    receiver = ReferenceString()


#
# Resource config
#
class RequestsResourceConfig(RecordResourceConfig):
    """Requests resource configuration."""

    blueprint_name = "requests"
    url_prefix = "/requests"
    routes = {
        "list": "/",
        "item": "/<id>",
        "action": "/<id>/actions/<action>",
    }

    request_view_args = {
        "id": ma.fields.Str(),
        "action": ma.fields.Str(),
    }

    request_search_args = RequestSearchRequestArgsSchema



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request Event Schemas."""

from datetime import timezone

from invenio_records_resources.services.records.schema import BaseRecordSchema
from marshmallow import (
    RAISE,
    Schema,
    ValidationError,
    fields,
    missing,
    validate,
    validates_schema,
)
from marshmallow_oneofschema import OneOfSchema
from marshmallow_utils import fields as utils_fields

from ..records.api import Request, RequestEventFormat, RequestEventType


class EntityReferenceBaseSchema(Schema):
    """Base schema for entity references, allowing only a single key.

    It will be populated dynamically by the ``RequestType``, based on the allowed
    reference types registered there.
    """

    class Meta:
        """Schema meta."""

        unknown = RAISE

    @validates_schema
    def there_can_be_only_one(self, data, **kwargs):
        """Only allow a single key."""
        if len(data) != 1:
            raise ValidationError("Entity references may only have one key")

    @classmethod
    def create_from_dict(cls, allowed_types, special_fields=None):
        """Create an entity reference schema based on the allowed reference types.

        Per default, a ``fields.String()`` field is registered for each of the type
        names in the ``allowed_types`` list.
        The field type can be customized by providing an entry in the
        ``special_fields`` dict, with the type name as key and the field type as value
         (e.g. ``{"user": fields.Integer()}``).
        """
        field_types = special_fields or {}
        for ref_type in allowed_types:
            # each type would be a String field per default
            field_types.setdefault(ref_type, fields.String())

        return cls.from_dict(
            {ref_type: field_types[ref_type] for ref_type in allowed_types}
        )


class RequestSchema(BaseRecordSchema):
    """Schema for requests.

    Note that the payload schema and the entity reference schemas (i.e. creator,
    receiver, and topic) are dynamically constructed and injected into this schema.
    """

    number = fields.String(dump_only=True)
    type = fields.String()
    title = utils_fields.SanitizedUnicode(default="")
    description = utils_fields.SanitizedUnicode()

    # status information is also likely set by the service
    status = fields.String(dump_only=True)
    is_closed = fields.Boolean(dump_only=True)
    is_open = fields.Boolean(dump_only=True)
    expires_at = utils_fields.TZDateTime(
        timezone=timezone.utc, format="iso", dump_only=True
    )
    is_expired = fields.Boolean(dump_only=True)

    class Meta:
        """Schema meta."""

        unknown = RAISE


class GenericRequestSchema(RequestSchema):
    """Generic request schema.

    CAUTION: This schema should not be used for the final validation of input
    data. Use the request type's own defined schema instead.

    This schema can be used in situations where you need to do basic validation
    or dumping of a request without the payload.

    This is used e.g. in Invenio-RDM-Records for dumping a request without
    having to know the specific request type.
    """

    created_by = fields.Dict()
    receiver = fields.Dict()
    topic = fields.Dict()


class ModelFieldStr(fields.Str):
    """Marshmallow field for serializing by attribute before item."""

    def get_value(self, obj, attr, **kwargs):
        """Return obj.attr or obj[attr] in this precedence order."""
        return getattr(obj, attr, obj.get(attr, missing))


class BaseEventSchema(BaseRecordSchema):
    """Base Event schema that other schemas should inherit from."""

    type = ModelFieldStr(required=True)
    created_by = fields.Dict(dump_only=True)


class CommentSchema(BaseEventSchema):
    """Comment schema."""

    class CommentExtra(Schema):
        """Comment extras schema."""

        content = utils_fields.SanitizedHTML(
            required=True, validate=validate.Length(min=1)
        )
        format = fields.Str(
            validate=validate.OneOf(choices=[e.value for e in RequestEventFormat]),
            load_default=RequestEventFormat.HTML.value,
        )

    payload = fields.Nested(CommentExtra, required=True)


class NoExtrasSchema(BaseEventSchema):
    """No extras schema."""


class RequestEventSchema(BaseRecordSchema, OneOfSchema):
    """Schema."""

    # TODO: Make this schema hookable? how?

    type_schemas = {
        RequestEventType.COMMENT.value: CommentSchema,
        RequestEventType.ACCEPTED.value: NoExtrasSchema,
        RequestEventType.DECLINED.value: NoExtrasSchema,
        RequestEventType.CANCELLED.value: NoExtrasSchema,
        RequestEventType.REMOVED.value: NoExtrasSchema,
        RequestEventType.EXPIRED.value: NoExtrasSchema,
    }
    type_field_remove = False

    def get_obj_type(self, obj):
        """Return key of type_schemas to use given object to dump."""
        return obj.type



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfields for request records."""

from .calculated import ExpiredStateCalculatedField, StateCalculatedField
from .entity_reference import ReferencedEntityField
from .identity import IdentityField
from .request_type import RequestTypeField
from .status import RequestStatusField

__all__ = (
    "ExpiredStateCalculatedField",
    "IdentityField",
    "ReferencedEntityField",
    "RequestStatusField",
    "RequestTypeField",
    "StateCalculatedField",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver for records."""

from sqlalchemy.exc import StatementError
from sqlalchemy.orm.exc import NoResultFound

from ..base import EntityProxy, EntityResolver


class RecordProxy(EntityProxy):
    """Resolver proxy for a Record entity using the pid."""

    def __init__(self, ref_dict, record_cls):
        """Constructor.

        :param record_cls: The record class to use.
        """
        super().__init__(ref_dict)
        self.record_cls = record_cls

    def _resolve(self):
        """Resolve the Record from the proxy's reference dict."""
        pid_value = self._parse_ref_dict_id(self._ref_dict)
        return self.record_cls.pid.resolve(pid_value)

    def get_need(self):
        """Return None since Needs are not applicable to records."""
        return None


class RecordPKProxy(RecordProxy):
    """Resolver proxy for a Record entity using the UUID."""

    def _resolve(self):
        """Resolve the Record from the proxy's reference dict."""
        id_ = self._parse_ref_dict_id(self._ref_dict)
        try:
            return self.record_cls.get_record(id_)
        except StatementError as exc:
            raise NoResultFound() from exc


class RecordResolver(EntityResolver):
    """Resolver for records."""

    def __init__(self, record_cls, type_key="record", proxy_cls=RecordProxy):
        """Constructor.

        :param record_cls: The record class to use.
        :param type_key: The value to use for the TYPE part of the reference dicts.
        """
        self.record_cls = record_cls
        self.type_key = type_key
        self.proxy_cls = proxy_cls

    def matches_entity(self, entity):
        """Check if the entity is a record."""
        return isinstance(entity, self.record_cls)

    def _reference_entity(self, entity):
        """Create a reference dict for the given record."""
        return {self.type_key: str(entity.pid.pid_value)}

    def matches_reference_dict(self, ref_dict):
        """Check if the reference dict references a request."""
        return self._parse_ref_dict_type(ref_dict) == self.type_key

    def _get_entity_proxy(self, ref_dict):
        """Return a RecordProxy for the given reference dict."""
        return self.proxy_cls(ref_dict, self.record_cls)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSONSchema directory."""

# TODO it'd be nice to re-use the #agent $ref, but that's defined in
#      rdm-records
#      -> move that $ref to a more generic package? records-resources maybe?



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Base class for entity resolvers."""

from abc import ABC, abstractmethod


def _parse_ref_dict(reference_dict, strict=True):
    """Parse the referenced dict into a tuple (TYPE, ID).

    The `strict` parameter controls if the number of keys in the reference dict
    is checked strictly or not.
    """
    keys = list(reference_dict.keys())

    if strict and len(keys) != 1:
        raise ValueError(
            "Reference dicts may only have one property! "
            f"Offending dict: {reference_dict}"
        )

    if not keys:
        return None

    type_ = keys[0]
    id_ = reference_dict[type_]
    return (type_, id_)


class EntityProxy(ABC):
    """Proxy for a type of entities which only resolves the entity if requested.

    EntityProxies are responsible for resolving the entities referenced in the
    reference dicts, and thus act as another layer of abstraction.
    They provide a unified interface for resolving the actual entities from the
    dicts, or creating Needs from the references (which is useful for permission
    checks, for instance on owners of records or requests).
    Also, they enable the actual entity to be lazily loaded via the explicit
    `resolve()` operation, which prevents unnecessary lookups if the entity is
    not of importance (e.g. when only the Need is relevant).
    After the first lookup, the resolved entity will be cached by the proxy object.
    """

    def __init__(self, reference_dict):
        """Constructor."""
        self._ref_dict = reference_dict
        self._entity = None

    def __repr__(self):
        """Return repr(self)."""
        return f"<{type(self).__name__} {self._ref_dict} ({self._entity})>"

    def _parse_ref_dict(self, ref_dict):
        """Parse the referenced dict into a tuple (TYPE, ID)."""
        return _parse_ref_dict(ref_dict)

    def _parse_ref_dict_type(self, ref_dict):
        """Parse the TYPE from the reference dict."""
        return self._parse_ref_dict(ref_dict)[0]

    def _parse_ref_dict_id(self, ref_dict):
        """Parse the ID from the reference dict."""
        return self._parse_ref_dict(ref_dict)[1]

    @property
    def reference_dict(self):
        """Get the proxy's reference dict."""
        return self._ref_dict

    def resolve(self):
        """Resolve the referenced entity via a query."""
        if self._entity is not None:
            # caching
            return self._entity

        self._entity = self._resolve()
        return self._entity

    @abstractmethod
    def _resolve(self):
        """The logic for performing the actual resolve operation.

        This method must be implemented by concrete subclasses.
        Note: The caching logic is already handled in `resolve()`, so this method
        should contain the pure resolution logic.
        """
        return None

    @abstractmethod
    def get_need(self):
        """Get the Need for the referenced entity, if applicable.

        If the concept is not applicable for this resolver's type of entities,
        `None` will be returned.
        """
        return None


class EntityResolver(ABC):
    """Translation layer between reference dicts and entities (or proxies).

    The resolvers are a layer of abstraction between the reference dicts and their
    referenced entities.
    They are directly responsible for handling the reference dumping direction
    themselves (i.e. creating reference dicts from entities), while the other direction
    (i.e. resolving entities from reference dicts) is the responsibility of the
    EntityProxies created by the resolvers.
    """

    def _parse_ref_dict(self, ref_dict):
        """Parse the referenced dict into a tuple (TYPE, ID)."""
        return _parse_ref_dict(ref_dict)

    def _parse_ref_dict_type(self, ref_dict):
        """Parse the TYPE from the reference dict."""
        return self._parse_ref_dict(ref_dict)[0]

    def _parse_ref_dict_id(self, ref_dict):
        """Parse the ID from the reference dict."""
        return self._parse_ref_dict(ref_dict)[1]

    def get_entity_proxy(self, ref_dict, check=True):
        """Check compatibility and get a proxy for the entity referenced in the dict.

        This method will optionally check the shape of the given reference dict via
        `matches_reference_dict()` before creating a proxy via `_get_entity_proxy()`.
        If this check fails, a `ValueError` will be raised.
        """
        if check and not self.matches_reference_dict(ref_dict):
            raise ValueError(
                f"{type(self).__name__} cannot handle "
                f"the following reference dict: {ref_dict}"
            )

        return self._get_entity_proxy(ref_dict)

    def reference_entity(self, entity, check=True):
        """Check compatibility and create a reference dict for the given entity.

        This method will perform an optional instance check of the given entity via
        `matches_entity()` before dumping a reference via `_reference_entity()`.
        If this check fails, a `ValueError` will be raised.
        """
        if check and not self.matches_entity(entity):
            raise ValueError(
                f"{type(self).__name__} cannot handle "
                f"the following entity: {entity}"
            )

        return self._reference_entity(entity)

    @abstractmethod
    def matches_reference_dict(self, ref_dict):
        """Check if the given ref_dict matches the expectations of this resolver."""
        return False

    @abstractmethod
    def matches_entity(self, entity):
        """Check if the given entity matches the expectations of this resolver."""
        return False

    @abstractmethod
    def _get_entity_proxy(self, ref_dict):
        """The logic for building a proxy for the referenced entity.

        Since the compatibility checks are already taken care of, this method can
        assume that the data is valid and can focus on simply creating the proxy
        for the referenced entity.
        """
        return None

    @abstractmethod
    def _reference_entity(self, entity):
        """The logic building a reference dict from the given entity.

        Since the compatibility checks are already taken care of, this method can
        assume that the entity is compatible and can focus on simply creating the
        reference dict for the given entity.
        """
        return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Default entity resolvers and proxies."""

from .records import RecordResolver
from .requests import RequestResolver
from .users import UserResolver

__all__ = (
    "RecordResolver",
    "RequestResolver",
    "UserResolver",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Registry for easy access to the registered entity resolvers."""

from invenio_access.permissions import system_process

from ..proxies import current_requests


class ResolverRegistry:
    """Namespace for resolver functions."""

    @classmethod
    def get_registered_resolvers(cls):
        """Get all currently registered resolvers."""
        return iter(current_requests.entity_resolvers_registry)

    @classmethod
    def resolve_entity_proxy(cls, reference_dict, raise_=False):
        """Get a proxy for the referenced entity via the configured resolvers.

        If `REQUESTS_ENTITY_RESOLVERS` does not contain a matching EntityResolver
        for the given `reference_dict`, the `raise_` parameter determines whether
        a `ValueError` is raised or `None` is returned.
        """
        for resolver in cls.get_registered_resolvers():
            if resolver.matches_reference_dict(reference_dict):
                return resolver.get_entity_proxy(reference_dict, check=False)

        if raise_:
            raise ValueError(f"No matching resolver registered for: {reference_dict}")

        return None

    @classmethod
    def resolve_entity(cls, reference_dict, raise_=False):
        """Resolve the referenced entity via the configured resolvers.

        This calls `resolve_entity_proxy()` and then `resolve()` on the proxy.
        Note that this may create an expensive lookup, like a DB query.
        """
        proxy = cls.resolve_entity_proxy(reference_dict, raise_=raise_)

        if proxy is None:
            return None

        return proxy.resolve()

    @classmethod
    def resolve_need(cls, reference_dict, raise_=False):
        """Get the need for the referenced entity via the configured resolvers.

        This calls `resolve_entity_proxy()` and then `get_need()` on the proxy.
        Note: If the concept of needs is not applicable to the referenced type of
        entity (e.g. `Request`), `None` is returned.
        """
        proxy = cls.resolve_entity_proxy(reference_dict, raise_=raise_)

        if proxy is None:
            return None

        return proxy.get_need()

    @classmethod
    def reference_entity(cls, entity, raise_=False):
        """Create a reference dict for the given entity via the configured resolvers.

        If `REQUESTS_ENTITY_RESOLVERS` does not contain a matching EntityResolver
        for the given `entity`, the `raise_` parameter determines whether a `ValueError`
        is raised or `None` is returned.
        """
        for resolver in cls.get_registered_resolvers():
            if resolver.matches_entity(entity):
                return resolver.reference_entity(entity, check=False)

        if raise_:
            raise ValueError(f"No matching resolver registered for: {type(entity)}")

        return None

    @classmethod
    def reference_identity(cls, identity, raise_=False):
        """Create a reference dict for the user behind the given identity."""
        if system_process in identity.provides:
            return None

        return {"user": str(identity.id)}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfields for calculated properties."""


from .base import CalculatedField
from .expired import ExpiredStateCalculatedField
from .state import StateCalculatedField

__all__ = (
    "CalculatedField",
    "ExpiredStateCalculatedField",
    "StateCalculatedField",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for managing referenced entities in request."""

from functools import partial

from invenio_records.systemfields import SystemField

from ...resolvers.base import EntityProxy
from ...resolvers.registry import ResolverRegistry


class ReferencedEntityField(SystemField):
    """Systemfield for managing the request type."""

    def __init__(self, key=None, reference_check_func=None):
        """Constructor."""
        super().__init__(key=key)
        self._ref_check = reference_check_func

    def _check_reference(self, instance, ref_dict):
        """Check if the reference is accepted."""
        if self._ref_check is None:
            return True

        return self._ref_check(instance, ref_dict)

    def set_obj(self, instance, obj):
        """Set the referenced entity."""
        # allow the setter to be used with a reference dict,
        # an entity proxy, or an actual entity
        if not isinstance(obj, dict):
            if isinstance(obj, EntityProxy):
                obj = obj.reference_dict
            elif obj is not None:
                obj = ResolverRegistry.reference_entity(obj, raise_=True)

        # check if the reference is allowed
        if not self._check_reference(instance, obj):
            raise ValueError(f"Invalid reference for '{self.key}': {obj}")

        # set dictionary key and reset the cache
        self.set_dictkey(instance, obj)
        self._set_cache(instance, None)

    def __set__(self, record, value):
        """Set the referenced entity."""
        assert record is not None
        self.set_obj(record, value)

    def obj(self, instance):
        """Get the referenced entity as an `EntityProxy`."""
        obj = self._get_cache(instance)
        if obj is not None:
            return obj

        reference_dict = self.get_dictkey(instance)
        if reference_dict is None:
            # TODO maybe use a `NullProxy` instead?
            return None

        obj = ResolverRegistry.resolve_entity_proxy(reference_dict)
        self._set_cache(instance, obj)
        return obj

    def __get__(self, record, owner=None):
        """Get the referenced entity as an `EntityProxy`."""
        if record is None:
            # access by class
            return self

        return self.obj(record)


def check_allowed_references(get_allows_none, get_allowed_types, request, ref_dict):
    """Check the reference according to rules specific to requests.

    In case the ``ref_dict`` is ``None``, it will check if this is allowed for the
    reference at hand.
    Otherwise, it will check if the reference dict's key (i.e. the TYPE) is allowed.
    """
    if ref_dict is None:
        return get_allows_none(request)

    ref_type = list(ref_dict.keys())[0]
    return ref_type in get_allowed_types(request)


check_allowed_creators = partial(
    check_allowed_references,
    lambda r: r.type.creator_can_be_none,
    lambda r: r.type.allowed_creator_ref_types,
)
"""Check function specific for the ``created_by`` field of requests."""

check_allowed_receivers = partial(
    check_allowed_references,
    lambda r: r.type.receiver_can_be_none,
    lambda r: r.type.allowed_receiver_ref_types,
)
"""Check function specific for the ``receiver`` field of requests."""

check_allowed_topics = partial(
    check_allowed_references,
    lambda r: r.type.topic_can_be_none,
    lambda r: r.type.allowed_topic_ref_types,
)
"""Check function specific for the ``topic`` field of requests."""
